// Receives and displays humidity and temperature from AcuRite wireless sensors
// Written by Boomer48: April 2019

//#define DEBUG
//#define Dual_Weather

// Input port for RF receiver data.
// Can be any of the Arduino defined digital pins
#define Pulse_In 2

byte RF_Bit_Count = 8;
byte RF_Byte_Count = 0;
byte RF_Byte = 0;
char ASCII[] = "0123456789ABCDEF";
byte RF_Message[] = {0, 0, 0, 0, 0, 0, 0};
unsigned long Start_Time = 0;
int Pulse_Width;
byte Started = false;
byte Synced = false;
byte Sync_Count = 0;
byte RF_Msg_ID = 0;
byte temp;

// Make routines globally visible
void RF_setup();
void RF_loop();

// Code goes here
void RF_setup() {
  digitalWrite(Pulse_In, INPUT_PULLUP); //turn on pullup

  // Calculate register settings for interrupt on change
#if ((Pulse_In <= 13) & (Pulse_In >= 8))
  temp = Pulse_In - 8;
  bitSet(PCMSK0, temp);
  PCICR = B00000001; // enable PCIE0 group
#else
  bitSet(PCMSK2, Pulse_In);
  PCICR = B00000100; // enable PCIE2 group
#endif
  PCIFR = B00000000; // clear all pin change interrupt flags
}

void RF_loop() {
  byte i;
  byte cksum;
  String junk;
  word temp_C;
  float Celsius;
  float Fahrenheit;

  if (RF_Byte_Count == 7) {
    // AcuRite sensor message is 7 bytes long
    noInterrupts();
    i = 0;
    cksum = 0;

    while (i <= 5) {
      cksum = cksum + RF_Message[i];
      i++;
    }

    if (cksum == RF_Message[6]) { // valid message

#ifdef DEBUG
      // ---------- Debug code that dumps raw Message bytes ----------
      i = 0;
      while (i < 7) {
        Send_Byte(ASCII[RF_Message[i] >> 4]); // display upper nibble first
        Send_Byte(ASCII[RF_Message[i] & 0x0F]);
        i++;
      }
      goto Cleanup;
#endif

      if (RF_Msg_ID == 0)
        RF_Msg_ID = RF_Message[0];

#ifdef Dual_Weather
      if (RF_Msg_ID == RF_Message[0])
        LCD_Line1();
      else
        LCD_Line2();
#endif

      // Check battery status
      if (RF_Message[2] != 0x44) {
        Send_Byte('B');
        Send_Byte('A');
        Send_Byte('T');
        Send_Byte('T');
        Send_Byte('E');
        Send_Byte('R');
        Send_Byte('Y');
        Send_Byte(' ');
        Send_Byte('L');
        Send_Byte('O');
        Send_Byte('W');
        Send_Byte(' ');
        Send_Byte('S');
        Send_Byte('E');
        Send_Byte('N');

        if (RF_Msg_ID == RF_Message[0])
          Send_Byte('1');
        else
          Send_Byte('2');

        goto Cleanup;
      }

      // Humidity
      junk = String(RF_Message[3] & 0x7F);
      Send_Byte(junk[0]);
      Send_Byte(junk[1]);
      Send_Byte('%');
      Send_Byte(' ');

      // Concatenate lower 4 bits of byte 5 with lower 7 bits of byte 6.
      // Left shift of byte 6 eliminates the parity bit and allows the
      // right shift of the 16-bit word to complete the bit alignment.
      temp_C = word(RF_Message[4] & 0x0F, (RF_Message[5] << 1));
      temp_C = temp_C >> 1;

      //Scale Celsius
      Celsius = (float(temp_C) / 10) - 100;
      junk = String(Celsius);

      // Celsius string length varies: (X.X, XX.X, -X.X, or -XX.X)
      // Only one character displayed after the decimal point
      i = 0;
      while (junk[i] != '.') {
        Send_Byte(junk[i]);
        i++;
      }

      Send_Byte(junk[i]); // decimal point
      Send_Byte(junk[i + 1]); // decimal digit

      Send_Byte('C');
      Send_Byte(' ');

      // Convert to Fahrenheit
      Fahrenheit = (Celsius * 1.8) + 32;
      junk = String(Fahrenheit);

      // Fahrenheit string length varies: (X.X, XX.X, XXX.X, -X.X, or -XX.X)
      // Only one character displayed after the decimal point
      i = 0;
      while (junk[i] != '.') {
        Send_Byte(junk[i]);
        i++;
      }

      Send_Byte(junk[i]); // decimal point
      Send_Byte(junk[i + 1]); // decimal digit

      Send_Byte('F');
      Send_Byte(' ');
      Send_Byte(' ');
      Send_Byte(' ');
    }

Cleanup:
    RF_Byte_Count = 0;
    Started = false;
    Synced = false;
    Sync_Count = 0;
    PCIFR = B00000000; // clear all pin change interrupt flags
    interrupts();
  }
} // end loop

// Interrupt-on-change handler
#if ((Pulse_In <= 13) & (Pulse_In >= 8))
ISR (PCINT0_vect) {
#else
ISR (PCINT2_vect) {
#endif
  //when the pin goes LOW record the pulse start time
  if (digitalRead(Pulse_In) == LOW) {
    Start_Time = micros();
  }
  else // pin went high
    if (Start_Time != 0) {
      //calculate the pulse width
      Pulse_Width = ((volatile int)micros() - Start_Time);
      //clear the timer
      Start_Time = 0;

      // Now check for valid message
      // Look for a 2ms low start pulse followed by four
      // 600us sync pulses.  Bits are either 400us or 200us.
      // Reject anything greater than 2.4ms
      if (Pulse_Width < 2400) {
        if (Started == true) {
          if (Synced == true) {
            if (Pulse_Width > 400) {
              // bit = 0
              RF_Bit_Count--; // bits received MSB first
              bitClear(RF_Message[RF_Byte_Count], RF_Bit_Count);
              if (RF_Bit_Count == 0) {
                RF_Byte_Count++;
                RF_Bit_Count = 8;
              }
            }
            else if (Pulse_Width > 200) {
              // bit = 1
              RF_Bit_Count--; // bits received MSB first
              bitSet(RF_Message[RF_Byte_Count], RF_Bit_Count);
              if (RF_Bit_Count == 0) {
                RF_Byte_Count++;
                RF_Bit_Count = 8;
              }
            }
            else { // bad bit
              Started = false;
              Synced = false;
              Sync_Count = 0;
            }
          } // Synced
          else if (Pulse_Width > 600) {
            Sync_Count++;
            if (Sync_Count == 4) {
              Synced = true;
            }
          }
          else { //sync was disrupted
            Started = false;
            Synced = false;
            Sync_Count = 0;
          }
        } // Started
        // Sync pulse should be about 2ms
        else if (Pulse_Width > 1800) {
          // start of message
          Started = true;
        }
      }
      else { // invalid message
        Started = false;
        Synced = false;
        Sync_Count = 0;
      }
    }
}
