// Displays local and UTC time using GGA message from a GPS receiver
// Written by Boomer48: April 2019


//#define Dual_Clock // displays both local time and UTC time

#include <EEPROM.h>

#define Switch_12_24 11 // switch to display 12 or 24 hour time
#define Switch_offset 12 // switch to enter UTC offset set mode

char NMEA_Char;
byte GPS_Count = 0;
char Last_Char = ' ';
byte GGA_Msg = false;
byte Time_Count;
byte Time_Array[6];
byte Table_Index;
byte Table_Value;
byte ROM_Byte;
byte UTC_Offset;
int  Time_Zone;
byte UTC_Timeout;

const byte Valid_Offset = 0xA5;

const byte Time12_Table[] = {0x12, 0x01, 0x02, 0x03, 0x04, 0x05, \
                             0x06, 0x07, 0x08, 0x09, 0x10, 0x11, \
                             0x12, 0x01, 0x02, 0x03, 0x04, 0x05, \
                             0x06, 0x07, 0x08, 0x09, 0x10, 0x11, \
                             0x12, 0x01, 0x02, 0x03, 0x04, 0x05, \
                             0x06, 0x07, 0x08, 0x09, 0x10, 0x11, \
                             0x12, 0x01, 0x02, 0x03, 0x04, 0x05, \
                             0x06, 0x07, 0x08, 0x09, 0x10, 0x11
                            };

const byte Time24_Table[] = {0x12, 0x13, 0x14, 0x15, 0x16, 0x17, \
                             0x18, 0x19, 0x20, 0x21, 0x22, 0x23, \
                             0x00, 0x01, 0x02, 0x03, 0x04, 0x05, \
                             0x06, 0x07, 0x08, 0x09, 0x10, 0x11, \
                             0x12, 0x13, 0x14, 0x15, 0x16, 0x17, \
                             0x18, 0x19, 0x20, 0x21, 0x22, 0x23, \
                             0x00, 0x01, 0x02, 0x03, 0x04, 0x05, \
                             0x06, 0x07, 0x08, 0x09, 0x10, 0x11
                            };

const char AM_PM[] = {'P', 'P', 'P', 'P', 'P', 'P', \
                      'P', 'P', 'P', 'P', 'P', 'P', \
                      'A', 'A', 'A', 'A', 'A', 'A', \
                      'A', 'A', 'A', 'A', 'A', 'A', \
                      'P', 'P', 'P', 'P', 'P', 'P', \
                      'P', 'P', 'P', 'P', 'P', 'P', \
                      'A', 'A', 'A', 'A', 'A', 'A', \
                      'A', 'A', 'A', 'A', 'A', 'A'
                     };

// Make routines globally visible
void GPS_Clock_setup();
void GPS_Clock_loop();
void Check_Time();
void Check_Offset();
void Disp_Offset_Msg();
void Disp_Offset_Val();
void Set_Offset();
void Disp_UTC_Time();
void Disp_Local_Time();

// Code goes here
void GPS_Clock_setup() {
  pinMode(Switch_12_24, INPUT_PULLUP);
  pinMode(Switch_offset, INPUT_PULLUP);
  Check_Offset();
  Serial.begin(9600);
  Send_Byte('N');
  Send_Byte('O');
  Send_Byte(' ');
  Send_Byte('D');
  Send_Byte('A');
  Send_Byte('T');
  Send_Byte('A');
}

void GPS_Clock_loop() {
  if (digitalRead(Switch_offset) == LOW)
    Set_Offset();
  else
    Check_Time();
}

// Checks EEPROM for a valid UTC offset
void Check_Offset() {
  ROM_Byte = EEPROM.read(0);

  if (ROM_Byte == Valid_Offset) {
    UTC_Offset = EEPROM.read(1);
    Disp_Offset_Msg();

    // Convert offset to integer for time calculations
    Time_Zone = (UTC_Offset & 0x0F) + (bitRead(UTC_Offset, 4) * 10);
    if (bitRead(UTC_Offset, 7) == 1)
      Time_Zone = -Time_Zone;

    delay(1000);
  }
  else {
    Set_Offset();
  }
}

void Disp_Offset_Msg() {
  LCD_Line1();
  Send_Byte('U');
  Send_Byte('T');
  Send_Byte('C');
  Send_Byte(' ');
  Send_Byte('O');
  Send_Byte('F');
  Send_Byte('F');
  Send_Byte('S');
  Send_Byte('E');
  Send_Byte('T');
  Send_Byte(' ');
  Disp_Offset_Val();
}

// Displays the UTC offset value (-12 to +12)
void Disp_Offset_Val() {
  // Start the display at a specific LCD location
  if (bitRead(UTC_Offset, 7) == 1)
    LCD_Char_Loc(0x8B, '-');
  else
    LCD_Char_Loc(0x8B, '+');

  Send_Byte(((UTC_Offset >> 4) & 0x07) + 0x30);
  Send_Byte((UTC_Offset & 0x0F) + 0x30);
  Send_Byte(' ');
  Send_Byte(' ');
}

// Set the UTC offset.  This routine gets called if no valid
// offset is found at power on or if the Offset Switch is pressed.
// A short press on the switch changes the sign or value.  A long
// press (> 1 second) moves to the next function or exits setup.
void Set_Offset() {
  byte IRQ_Mask;

  IRQ_Mask = PCICR; // save pin change interrupt mask
  PCICR = 0; // disable pin change interrupts
  PCIFR = 0; // clear all pin change interrupt flags

  UTC_Offset = 0x77; // display out of range value
  Disp_Offset_Msg();
  UTC_Offset = 0x00;

  while (digitalRead(Switch_offset) == LOW) {
    delay(100); // wait for switch to be released
  }

  while (digitalRead(Switch_offset) == HIGH) {
    delay(100); // wait for switch to be pressed
  }

  UTC_Timeout = 0;

  while (digitalRead(Switch_offset) == LOW) {
    delay(100); // wait for switch to be released
    UTC_Timeout++;
  }

  // Toggle sign bit
  // Move on if switch press > 1 second
  while (UTC_Timeout < 10) {
    if (bitRead(UTC_Offset, 7) == 0)
      bitSet(UTC_Offset, 7);
    else
      bitClear(UTC_Offset, 7);

    Disp_Offset_Val();

    while (digitalRead(Switch_offset) == HIGH) {
      delay(100); // wait for switch to be pressed
    }

    UTC_Timeout = 0;

    while (digitalRead(Switch_offset) == LOW) {
      delay(100); // wait for switch to be released
      UTC_Timeout++;
    }
  }

  while (digitalRead(Switch_offset) == HIGH) {
    delay(100); // wait for switch to be pressed
  }

  UTC_Timeout = 0;

  // Skip offset increment if switch press > 1 second
  // Leaves value at zero offset
  while (digitalRead(Switch_offset) == LOW) {
    delay(100); // wait for switch to be released
    UTC_Timeout++;
  }

  // Increment offset value
  // Move on if switch press > 1 second
  while (UTC_Timeout < 10) {
    if ((UTC_Offset & 0x7F) < 0x12)
      UTC_Offset++;

    if ((UTC_Offset & 0x7F) == 0x0A)
      UTC_Offset = (UTC_Offset & 0xF0) + 0x10;

    Disp_Offset_Val();

    while (digitalRead(Switch_offset) == HIGH) {
      delay(100); // wait for switch to be pressed
    }

    UTC_Timeout = 0;

    while (digitalRead(Switch_offset) == LOW) {
      delay(100); // wait for switch to be released
      UTC_Timeout++;
    }
  }

  // Save UTC offset in EEPROM
  EEPROM.write(0, Valid_Offset);
  EEPROM.write(1, UTC_Offset);

  // Convert to integer for time calculations
  Time_Zone = (UTC_Offset & 0x0F) + (bitRead(UTC_Offset, 4) * 10);
  if (bitRead(UTC_Offset, 7) == 1)
    Time_Zone = -Time_Zone;

  LCD_Line1();
  Send_Byte('S');
  Send_Byte('A');
  Send_Byte('V');
  Send_Byte('E');
  Send_Byte('D');
  Send_Byte(' ');
  Send_Byte('O');
  Send_Byte('F');
  Send_Byte('F');
  Send_Byte('S');
  Send_Byte('E');
  Send_Byte('T');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  Send_Byte(' ');
  delay(1000);
  Serial.flush();
  PCICR = IRQ_Mask; // enable pin change interrupts
}

// Read GPS messages and discard all but GGA
// Save first 6 bytes after first comma in GGA message
void Check_Time() {
  if (Serial.available() > 0) {
    // read the incoming byte
    NMEA_Char = Serial.read();

    if (GGA_Msg == true) {
      if (Last_Char == ',') {
        // time processing
        if ((NMEA_Char >= '0') & (NMEA_Char <= '9')) {
          Time_Array[Time_Count] = NMEA_Char;
          Time_Count++;
          if (Time_Count == 6) {
            GGA_Msg = false;
#ifdef Dual_Clock
            Disp_UTC_Time();
#endif
            Disp_Local_Time();
          }
        }
      }
      else if (NMEA_Char == ',') {
        Last_Char = NMEA_Char;
        Time_Count = 0;
      }
      else {
        GGA_Msg = false;
      }
    }

    else if ((NMEA_Char == 'A') & (Last_Char == 'G'))
      GGA_Msg = true;
    else
      Last_Char = NMEA_Char;
  }
}

// Display local time (12-hour or 24-hour format)
// Requires conversion of UTC hours from ASCII to integer
// Requires time zone conversion (uses saved UTC offset)
// Requires conversion to 12-hour format if selected
void Disp_Local_Time() {
  LCD_Line1();
  // Start the index into Time_Array at 12 then add the UTC
  // offset(range: -12 to +12)
  Table_Index = (((Time_Array[0] & 0x0F) * 10) + \
                 ((Time_Array[1] & 0x0F)) + 12 + Time_Zone);

  // Check 12/24 hour switch
  if (digitalRead(Switch_12_24) == HIGH)
    Table_Value = Time12_Table[Table_Index];
  else
    Table_Value = Time24_Table[Table_Index];

  // Convert table value hours from hex to ASCII
  Time_Array[1] = (Table_Value & 0x0F) + 0x30;
  Time_Array[0] = ((Table_Value & 0xF0) >> 4) + 0x30;

  Send_Byte(' ');
  Send_Byte(' ');

  // Suppress leading zero if in 12-hour mode
  if (digitalRead(Switch_12_24) == HIGH)
    if (Time_Array[0] != '0')
      Send_Byte(Time_Array[0]);
    else
      Send_Byte(' ');
  else
    Send_Byte(Time_Array[0]);

  Send_Byte(Time_Array[1]);
  Send_Byte(':');
  Send_Byte(Time_Array[2]);
  Send_Byte(Time_Array[3]);
  Send_Byte(':');
  Send_Byte(Time_Array[4]);
  Send_Byte(Time_Array[5]);
  Send_Byte(' ');
  Send_Byte(' ');

  // Add AM/PM if in 12-hour mode
  if (digitalRead(Switch_12_24) == HIGH) {
    Table_Value = AM_PM[Table_Index];
    Send_Byte(Table_Value);
    Send_Byte('M');
  }
  else {
    Send_Byte(' ');
    Send_Byte(' ');
  }

  Send_Byte(' ');
  Send_Byte(' ');
}

// Display the ASCII time characters from the GGA message
void Disp_UTC_Time() {
  // Display UTC Time
  LCD_Line2();

  Send_Byte(' ');
  Send_Byte(' ');

  Send_Byte(Time_Array[0]);
  Send_Byte(Time_Array[1]);
  Send_Byte(':');
  Send_Byte(Time_Array[2]);
  Send_Byte(Time_Array[3]);
  Send_Byte(':');
  Send_Byte(Time_Array[4]);
  Send_Byte(Time_Array[5]);
  Send_Byte(' ');
  Send_Byte('U');
  Send_Byte('T');
  Send_Byte('C');
  Send_Byte(' ');
  Send_Byte(' ');
}
